"""EPP Handler, carries the EPP session."""
import logging
from typing import Optional, cast

from epplib.client import Client
from epplib.commands import Login, Logout
from epplib.commands.base import Request as EppRequest, ResponseT
from epplib.exceptions import TransportError
from epplib.policy import Policy
from epplib.transport import SocketTransport

from .exceptions import EppicError, ServerError
from .settings import Session

LOGGER = logging.getLogger(__name__)


class EppHandler:
    """Wrapper around the epplib `Client`, handling session with the server.

    Note: The connection to the server is tied to the login session.
    """

    def __init__(self, session: Optional[Session]) -> None:
        self._client: Optional[Client] = None
        self._session: Optional[Session] = None

        if session is not None:
            self.session = session

    @property
    def client(self) -> Client:
        """Return an active `Client` instance.

        If a session with the server isn't set, create one.
        If the session information is invalid (hostname not set), raise an exception.
        """
        if self._client is None:
            if self.session.hostname is None:
                raise EppicError("Hostname not specified.")
            transport = SocketTransport(
                self.session.hostname, self.session.port,
                cert_file=self.session.cert_file,
                key_file=self.session.key_file,
                verify=self.session.verify
            )
            self._client = Client(transport)

        return self._client

    @property
    def logged_in(self) -> bool:
        """Return `True` if there is currently an active session with the server."""
        return self._client is not None

    @property
    def session(self) -> Session:
        """Get the current `Session` object.

        If session isn't set, raise an exception.
        """
        if self._session is None:
            raise EppicError("Session not set.")

        return self._session

    @session.setter
    def session(self, session: Session) -> None:
        """Set a session.

        Note: if there's a different active session, it gets closed properly.
        Note: if the current session has login information (username, password),
            login is performed automatically.
        """
        self.logout()

        self._session = session

        # allows defining sessions without username and password
        # so user can log in manually later
        if session.username and session.password:
            self.login()

    def logout(self) -> None:
        """Log out of an active session and close the connection to the server.

        If not logged in, do nothing.
        If the logout or termination of the connection is unsuccessful in any way, give a warning.
        """
        if not self.logged_in:
            return

        LOGGER.info("Logging out and closing connection...")
        try:
            self._send_request(Logout())
        except Exception:
            LOGGER.warning("Logout failed.")

        try:
            self.client.close()
        except Exception:
            LOGGER.warning("Connection failed on close.")

        self._client = None

    def login(self) -> None:
        """Send a `Login` command with information from the current `session`.

        Note: if there's a different active session, it gets closed properly.

        Raises an exception if the connection with the server fails or the login is unsuccessful.
        """
        self.logout()

        username = cast(str, self.session.username)
        password = cast(str, self.session.password)

        LOGGER.info("Logging in...")
        try:
            self.client.connect()
            response = self.send_request(Login(cl_id=username, password=password,
                                               obj_uris=self.session.obj_uris))
            if response.code >= 2000:
                if response.code in (2200, 2501):
                    raise ServerError("Invalid username or password.")
                else:
                    raise ServerError("Login was unsuccessful.")
        except Exception:
            self.logout()
            raise

    @property
    def policy(self) -> Optional[Policy]:
        """Get `Policy` of the client, or `None` if not logged in."""
        return self.client.policy if self.logged_in else None

    def send_request(self, request: EppRequest[ResponseT], cltrid: Optional[str] = None) -> ResponseT:
        """Send an EPP request to the server, return the server response.

        If sending the request fails, connection to the server is properly terminated and an exception is raised.
        """
        try:
            return self._send_request(request, cltrid)
        except TransportError as e:
            LOGGER.exception("Communication with the server failed. Logging out and closing connection...")
            self.logout()
            raise ServerError("Communication with the server failed.") from e

    def _send_request(self, request: EppRequest[ResponseT], cltrid: Optional[str] = None) -> ResponseT:
        """Send an EPP request to the server, return the server response."""
        LOGGER.info("Send request: %s", request)
        response = self.client.send(request, tr_id=cltrid)

        # store the raw XML, because the `client` can be killed (in an error, for example)
        self.last_raw_command = self.client.last_raw_command
        self.last_raw_response = self.client.last_raw_response

        return response
