"""Tests for Service Commands."""
from typing import Any, Dict, cast
from unittest import TestCase
from unittest.mock import patch

from eppic.commands.service_requests import (
    ServiceChangeSettings,
    ServiceExit,
    ServiceLogin,
    ServiceLogout,
    ServiceNewSession,
)
from eppic.output import JsonOutputHandler
from eppic.settings import OutputFormat, Session

from .utils import GetAppMixin


class TestServiceExit(GetAppMixin, TestCase):

    @patch("eppic.commands.service_requests.sys")
    def test_exit(self, sys):
        command = ServiceExit()
        command.handle(self.get_app())
        sys.exit.assert_called_once_with(0)


class TestServiceLogin(GetAppMixin, TestCase):

    def test_service_login_already_logged(self):
        command = ServiceLogin("asdff")
        command.handle(self.get_app())
        self.assertEqual(self.epp_handler.session.username, "asdff")

    def test_service_login_not_logged(self):
        command = ServiceLogin("asdff")
        command.handle(self.get_app(session=None))
        self.assertEqual(self.epp_handler.session.username, "asdff")

    def test_service_login_incorrect_session(self):
        self.assertRaises(ValueError, ServiceLogin("hello").handle, self.get_app())

    def test_service_login_no_sessions_exist(self):
        self.assertRaises(ValueError, ServiceLogin().handle, self.get_app(sessions=[], session=None))


class TestServiceLogout(GetAppMixin, TestCase):

    def test_service_logout(self):
        command = ServiceLogout()
        command.handle(self.get_app())
        self.epp_handler.logout.assert_called()

    def test_service_logout_not_logged_in(self):
        command = ServiceLogout()
        command.handle(self.get_app(session=None))  # doesn't throw an exception
        self.epp_handler.logout.assert_called()


class TestServiceNewSession(GetAppMixin, TestCase):

    def test_service_modify_some(self):
        command = ServiceNewSession(base_session_id="asdff", username="asdfgh")
        command.handle(self.get_app())
        self.assertEqual(self.epp_handler.session.username, "asdfgh")
        self.assertEqual(self.epp_handler.session.hostname, "asdf")

    def test_service_modify_none(self):
        command = ServiceNewSession(base_session_id="asdff")
        command.handle(self.get_app())
        self.assertEqual(self.epp_handler.session.username, "asdff")
        self.assertEqual(self.epp_handler.session.hostname, "asdf")

    def test_service_all_new(self):
        command = ServiceNewSession(hostname="asdfgh", port=123)
        command.handle(self.get_app())
        session = Session(hostname="asdfgh", port=123)
        self.assertEqual(self.epp_handler.session, session)
        self.assertEqual(self.epp_handler.session, session)

    def test_service_new_session_invalid_id(self):
        self.assertRaises(ValueError,
                          ServiceNewSession(base_session_id="hello there", hostname="asdfgh", port=123).handle,
                          self.get_app())


class TestServiceChangeSettings(GetAppMixin, TestCase):

    def test_service_change_settings(self):
        tests = [  # command input, test
            ({"output_format": OutputFormat.JSON}, lambda a: isinstance(a.output_handler, JsonOutputHandler)),
            ({"autocomplete": False}, lambda a: a.input_handler.prompt.completer is None),
            ({"syntax_highlighting": False}, lambda a: a.input_handler.prompt.lexer is None),
            ({"command_confirmation": True}, lambda a: a.settings.interactive.command_confirmation is True),
            ({"print_raw_xml": True}, lambda a: a.settings.print_raw_xml is True),
        ]
        for kwargs, test in tests:
            kwargs = cast(Dict[str, Any], kwargs)
            with self.subTest(**kwargs):
                app = self.get_app(autocomplete=True)
                self.assertFalse(test(app))
                ServiceChangeSettings(**kwargs).handle(app)
                self.assertTrue(test(app))
