import logging
import tempfile
from pathlib import Path
from typing import Any, Dict
from unittest import TestCase
from unittest.mock import patch

import yaml
from click.testing import CliRunner

from eppic.main import main
from eppic.settings import Session


class TestMain(TestCase):

    def setUp(self) -> None:
        self.tmp_dir = tempfile.TemporaryDirectory()
        self.addCleanup(self.tmp_dir.cleanup)
        self.tmp_file = Path(self.tmp_dir.name) / "settings.yaml"
        with open(Path(self.tmp_dir.name) / "key", "w") as f:
            f.write("foo")

        with open(Path(self.tmp_dir.name) / "cert", "w") as f:
            f.write("foo")

        self.content: Dict[str, Any] = {
            "sessions": [
                {
                    "hostname": "asdf",
                    "port": 1234,
                    "cert_file": str(Path(self.tmp_dir.name) / "cert"),
                    "key_file": str(Path(self.tmp_dir.name) / "key"),
                    "obj_uris": ["asdf"],
                    "username": "asdf",
                    "password": "ghjkl"
                }
            ],
        }
        self.write_settings(self.content)

    def run_eppic_main(self, args="", tmp_config=True):
        """Invoke the `main` function.

        Returns a tuple:
            The `click` result.
            A `Mock` object representing `Application`.
        """
        args = (f"--config {self.tmp_file} " + args) if tmp_config else args
        with patch("eppic.main.Application") as app:
            result = CliRunner().invoke(main, args)
        return result, app

    def assertResult(self, result, app, *, exit_code=0, new_session=None):
        """Test the result of invoking main.

        Tests exit code.
        Tests `ServiceNewSession` calls.
        """
        self.assertEqual(result.exit_code, exit_code)
        if exit_code == 0:
            app.assert_called_once()

        if new_session is None:
            app.return_value.add_session.assert_not_called()
        else:
            app.return_value.add_session.assert_called_once_with(
                Session(**new_session)
            )
            self.assertIsNone(self.get_app_args(app)["session"])

    def get_app_args(self, app):
        app.assert_called_once()
        args = app.call_args.args
        return {"settings": args[0], "command": args[1], "session": args[2]}

    def write_settings(self, content: Any) -> None:
        with open(self.tmp_file, "w") as f:
            yaml.safe_dump(content, f)

    def tearDown(self) -> None:
        self.tmp_dir.cleanup()

    def test_settings_error(self):
        (Path(self.tmp_dir.name) / "cert").rename(Path(self.tmp_dir.name) / "crt")
        result, app = self.run_eppic_main()
        self.assertResult(result, app, exit_code=1)
        app.assert_not_called()

    def test_main_run(self):
        result, app = self.run_eppic_main()
        self.assertResult(result, app)
        self.assertIsNone(self.get_app_args(app)["command"])

    def test_main_w_command(self):
        result, app = self.run_eppic_main("asdf sadf asdf")
        self.assertResult(result, app)
        self.assertEqual(self.get_app_args(app)["command"], "asdf sadf asdf")

    def test_main_w_input(self):
        result, app = self.run_eppic_main(f"--input-file {self.tmp_file}")
        self.assertResult(result, app)
        self.assertEqual(self.get_app_args(app)["settings"].input_file, self.tmp_file)

    def test_main_w_logging_set(self):
        logging.basicConfig(level="INFO")
        self.content["logging"] = {"version": 1, "root": {"level": "DEBUG"}}
        self.write_settings(self.content)
        result, app = self.run_eppic_main()
        self.assertResult(result, app)
        self.assertEqual(logging.getLogger().level, logging.DEBUG)

    def test_main_new_session(self):
        test_cases = [
            # session id is unused
            ("--session-id test --hostname hello --port 123 --username foobar", {"hostname": "hello",
                                                                                 "port": 123,
                                                                                 "username": "foobar"}),
            # include all fields
            ("--hostname hello --port 123 --username foobar --password passwd\
                    --obj-uris asdf --obj-uris ghjkl --alias foobarbaz", {"hostname": "hello",
                                                                          "port": 123,
                                                                          "username": "foobar",
                                                                          "password": "passwd",
                                                                          "obj_uris": ["asdf", "ghjkl"],
                                                                          "alias": "foobarbaz"}),
            # minimal
            ("--hostname hello", {"hostname": "hello"}),
            # minimal - obj uris
            ("--obj-uris foo", {"obj_uris": ["foo"]})
        ]
        for inp, expected_kwargs in test_cases:
            with self.subTest(input_args=inp):
                result, app = self.run_eppic_main(inp)
                self.assertResult(result, app, new_session=expected_kwargs)
